/**
 ** sparkSituationAssessmentCodels.cc
 **
 ** Codels called by execution task sparkMainTask
 **
 ** Author: E. Akin Sisbot
 ** Date: Aug 10
 **
 **/

#include <portLib.h>

#include "server/sparkHeader.h"
#include "sparkm3d.h"

#ifdef ORO
#include "sparkOro.h"
#endif

static SPARK_ALL_REACHABILITIES *gAllReachs = NULL;
static SPARK_ALL_OBJECT_STATES *gPrevObjectStates = NULL;

/* ----------------- INTERNAL FUNCTIONS -------------------- */

static int object_visibility_state(p3d_rob *object, HRI_AGENT *agent, 
				   double elevation, double azimuth,
				   int visThr,
				   SPARK_VISIBILITY *visibility)
{
  configPt sourceQ, savedQ;
  p3d_rob * robot = NULL;
  p3d_jnt *panJnt = NULL, *tiltJnt = NULL;
  int visResult;
  
  // MOVE THE AGENT TO LOOK AT THE TARGET
  
  robot  = agent->robotPt;
  panJnt  = robot->joints[agent->perspective->pan_jnt_idx];
  tiltJnt = robot->joints[agent->perspective->tilt_jnt_idx];
 
  
  if( (robot==NULL) || (panJnt==NULL) || (tiltJnt==NULL) ) {
    printf("In %s:%d, bad input\n",__FILE__,__LINE__);
    return FALSE;
  }
  
  sourceQ = MY_ALLOC(double, robot->nb_dof); /* ALLOC */
  savedQ  = MY_ALLOC(double, robot->nb_dof); /* ALLOC */
  
  p3d_get_robot_config_into(robot, &sourceQ);
  p3d_get_robot_config_into(robot, &savedQ);
  
  /* TURN PAN */
  azimuth = azimuth + sourceQ[panJnt->index_dof]; // Relative to absolute angle

  if(panJnt->dof_data[0].vmin > azimuth) {
    if(panJnt->dof_data[0].vmax < azimuth - M_2PI) {
      if(ABS(panJnt->dof_data[0].vmin-azimuth) < ABS(panJnt->dof_data[0].vmax-azimuth)) {
	sourceQ[panJnt->index_dof] = panJnt->dof_data[0].vmin;
      }
      else{
	sourceQ[panJnt->index_dof] = panJnt->dof_data[0].vmax;
      } 
    }
    else{
      sourceQ[panJnt->index_dof] = azimuth - M_2PI; 
    }	
  }
  else{
    if(panJnt->dof_data[0].vmax < azimuth) {
      if(panJnt->dof_data[0].vmin > azimuth - M_2PI) {
	if(ABS(panJnt->dof_data[0].vmin-azimuth) < ABS(panJnt->dof_data[0].vmax-azimuth)) {
	  sourceQ[panJnt->index_dof] = panJnt->dof_data[0].vmin;
	}
	else{
	  sourceQ[panJnt->index_dof] = panJnt->dof_data[0].vmax;
	}
      } 
      else{
	sourceQ[panJnt->index_dof] = azimuth- M_2PI; 
      }	
    }
    else{
      sourceQ[panJnt->index_dof] = azimuth; 
    }
  } 
  
  /* TURN TILT */
  elevation = elevation + sourceQ[tiltJnt->index_dof]; // Relative to absolute angle
 
  if(tiltJnt->dof_data[0].vmin > elevation) {
    sourceQ[tiltJnt->index_dof] = tiltJnt->dof_data[0].vmin;
  }
  else{
    if(tiltJnt->dof_data[0].vmax < elevation) {
      sourceQ[tiltJnt->index_dof] = tiltJnt->dof_data[0].vmax;
    }
    else{
      sourceQ[tiltJnt->index_dof] = elevation;  
    }
  }
  
  p3d_set_and_update_this_robot_conf(robot, sourceQ);  
  
  // VISIBILITY TEST
  
  visResult = hri_is_object_visible(agent, object, visThr, FALSE, FALSE);

  switch(visResult) {
  case 0: // INVISIBLE
    *visibility = INVISIBLE;
    break;
  case 1: // VISIBLE
    *visibility = VISIBLE;
    break;
  }
  p3d_set_and_update_this_robot_conf(robot, savedQ);
  
  MY_FREE(savedQ, double, robot->nb_dof); /* FREE */
  MY_FREE(sourceQ, double, robot->nb_dof); /* FREE */

  return TRUE;
}

/* sparkIsVisibleMain  -  codel EXEC of IsVisible
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkIsVisibleMain(SPARK_VISREACH_QUERY *visQuery, SPARK_VISREACH_RESULT *visResult, int *report)
{
  p3d_rob * target = NULL;
  HRI_AGENT * sourceAgent = NULL;
  int i;
  double elevation = 0, azimuth = 0;
  SPARK_VISIBILITY visibility;
  int visibPl;


  // -----Performance evaluation
  //struct timeval tvBegin, tvEnd, tvDiff;
  //gettimeofday(&tvBegin, NULL);
  // -----

  if(gAgents == NULL) {
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }
  
  // Find the target object
  for(i=0; i<gEnv->nr; i++) {
    if( strcasestr(gEnv->robot[i]->name, visQuery->target.name) ) {
      target = gEnv->robot[i];
      break;
    }
  }
  if(target == NULL) {
    *report = S_spark_OBJECT_NOT_FOUND;
    return ETHER;
  }
  
  //Find the source agent
  for(i=0; i<gAgents->all_agents_no; i++) {
    if( strcasestr(gAgents->all_agents[i]->robotPt->name, visQuery->agent.name) ) {
      sourceAgent = gAgents->all_agents[i];
      break;
    }
  }
  if(sourceAgent == NULL) {
    *report = S_spark_AGENT_NOT_FOUND;
    return ETHER;
  }
  
  hri_object_visibility_placement(sourceAgent, target, &visibPl, &elevation, &azimuth);
  object_visibility_state(target, sourceAgent, elevation, azimuth, SPARK_DEFAULT_VISIBILITY_RATIO, &visibility);
  
  switch((SPARK_VISIBILITY_PLACEMENT) visibPl) {
  case FOA: //in FOA 
    strcpy(visResult->result1.name,"FOA");
    break;
  case FOV: //in FOV 
    strcpy(visResult->result1.name,"FOV");
    break;
  case OOF: //in OOF
    strcpy(visResult->result1.name,"OOF");
    break;
  default: //should never come here
    printf("Problem in visibility_placement\n");
    break;
  }
  
  switch(visibility) {
  case INVISIBLE:
    strcpy(visResult->result2.name,"INVISIBLE");
    break; 
  case VISIBLE:
    strcpy(visResult->result2.name,"VISIBLE");
    break;
  }

  // -----Performance evaluation
  //gettimeofday(&tvEnd, NULL);
  //timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
  //printf("%ld.%06ld\n", tvDiff.tv_sec, tvDiff.tv_usec);
  // -------
  
  *report = OK;
  return ETHER;
}

SPARK_REACHABILITY spark_object_reachability_state(p3d_rob * object, HRI_AGENT *agent)
{
  int reached = FALSE;
  configPt qs, q;
  p3d_vector3 Tcoord[3];
  SPARK_REACHABILITY reachability;

  // Target object
  Tcoord[0][0] = object->joints[1]->abs_pos[0][3];      
  Tcoord[0][1] = object->joints[1]->abs_pos[1][3];     
  Tcoord[0][2] = object->joints[1]->abs_pos[2][3];
  
  if(D3D(Tcoord[0][0],Tcoord[0][1],Tcoord[0][2],
	 agent->robotPt->joints[1]->abs_pos[0][3],
	 agent->robotPt->joints[1]->abs_pos[1][3],
	 agent->robotPt->joints[1]->abs_pos[2][3]) > 1.5) {
    
    reachability = UNREACHABLE;
  }

  GIK_VIS = 500;
  
  qs = MY_ALLOC(double, agent->robotPt->nb_dof); /* ALLOC */
  q  = MY_ALLOC(double, agent->robotPt->nb_dof); /* ALLOC */
  
  p3d_get_robot_config_into(agent->robotPt, &qs); /* Saving agent config */
  
  reached = hri_agent_single_task_manip_move(agent, GIK_LATREACH, object, &q);
  
  if(agent->manip->type == TWO_ARMED) {
    p3d_set_and_update_this_robot_conf(agent->robotPt, qs);
    reached = hri_agent_single_task_manip_move(agent, GIK_RATREACH, object, &q); 
  }
  p3d_set_and_update_this_robot_conf(agent->robotPt, q);
  p3d_col_activate_robot(agent->robotPt);
  
  // In the end of the motion check if there is a collision
  if(!p3d_col_test_robot(agent->robotPt, 0)) {
    // Check if the collision is with the object
    p3d_col_deactivate_robot_robot(agent->robotPt, object);  
    if(!p3d_col_test_robot(agent->robotPt, 0)) {
      // The robot colliding another object
      if(reached)
	reachability = REACHABLE_WITH_COLLISION;
      else
	reachability = UNREACHABLE;
    }
    else {
      // Robot collides only with the target
      // That means it is reached
      reachability = REACHABLE;
    }
    p3d_col_activate_robot_robot(agent->robotPt, object); 
  }
  else {
    if(reached) {      
      reachability = REACHABLE;
    }
    else {
      reachability = UNREACHABLE;
    }
  }

  p3d_set_and_update_this_robot_conf(agent->robotPt,qs);
  
  MY_FREE(q, double , agent->robotPt->nb_dof); /* FREE */
  MY_FREE(qs, double, agent->robotPt->nb_dof); /* FREE */
  
  return reachability;
}

/* ----------------- END INTERNAL FUNCTIONS ----------------- */

/* sparkIsReachableMain  -  codel EXEC of IsReachable
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkIsReachableMain(SPARK_VISREACH_QUERY *reachQuery, SPARK_VISREACH_RESULT *reachResult, int *report)
{
  int i;
  p3d_rob * object = NULL;
  HRI_AGENT * agent = NULL;
  SPARK_REACHABILITY reachability;
 
  
  if(gEnv == NULL || gAgents==NULL) {
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }
  
  for(i=0; i<gEnv->nr; i++) {
    if(strcasestr(gEnv->robot[i]->name, reachQuery->target.name)) {
      object = gEnv->robot[i];
      break;
    }
  }
  if(object == NULL) {
    *report = S_spark_OBJECT_NOT_FOUND;
    return ETHER;
  }

  for(i=0; i<gAgents->all_agents_no; i++) {
    if(strcasestr(gAgents->all_agents[i]->robotPt->name, reachQuery->agent.name)) {
      agent = gAgents->all_agents[i];
      break;
    }
  }
  if(agent == NULL) {
    *report = S_spark_AGENT_NOT_FOUND;
    return ETHER;
  }
  
  reachability = spark_object_reachability_state(object, agent);

  switch(reachability) {
  case REACHABLE: 
    strcpy(reachResult->result1.name,"Reachable");
    break;
  case REACHABLE_WITH_COLLISION:
    strcpy(reachResult->result1.name,"ReachableWithCollision");
    break;
  case UNREACHABLE:
    strcpy(reachResult->result1.name,"Unreachable");
    break;
  }

  *report = OK;
  return ETHER;
}



static SPARK_PLACEMENT_RELATION placement_relation(p3d_rob *sourceObj, p3d_rob *targetObj)
{
  double sourceObjCx, sourceObjCy, sourceObjCz, targetObjCx, targetObjCy, targetObjCz;
  int nextTo = FALSE;
  
  if(sourceObj == NULL || targetObj == NULL) {
    printf("%s:%d plrelation input is null",__FILE__,__LINE__);
    return NOPLR;
  }
  
  sourceObjCx = (sourceObj->BB.xmin + sourceObj->BB.xmax)/2;
  sourceObjCy = (sourceObj->BB.ymin + sourceObj->BB.ymax)/2;
  sourceObjCz = (sourceObj->BB.zmin + sourceObj->BB.zmax)/2;
  
  targetObjCx = (targetObj->BB.xmin + targetObj->BB.xmax)/2;
  targetObjCy = (targetObj->BB.ymin + targetObj->BB.ymax)/2;
  targetObjCz = (targetObj->BB.zmin + targetObj->BB.zmax)/2;
  
  /* Compute IN */
  
  /* Test if source Obj is in targetObj */
  /* Condition: Source Obj BB should be wholly in targetObj BB */ 
  
  if((targetObj->BB.xmin <= sourceObj->BB.xmin) && (targetObj->BB.xmax >= sourceObj->BB.xmax) &&
     (targetObj->BB.ymin <= sourceObj->BB.ymin) && (targetObj->BB.ymax >= sourceObj->BB.ymax) &&
     (targetObj->BB.zmin <= sourceObj->BB.zmin) && (targetObj->BB.zmax >= sourceObj->BB.zmax)) {
    return ISIN;
  }
  
  /* Compute ON */
  
  /* Test if sourceObj is on targetObj */
  /* Condition 1: The center of SourceObj BB should be in the x,y limits of targetObj BB */
  /* Condition 2: The lower part of SourceObj BB should either intersect with the upper part of targetObj BB */ 
  /*              or there should be few cm's */ 
  
  
  if((sourceObjCx >= targetObj->BB.xmin) && (sourceObjCx <= targetObj->BB.xmax) &&
     (sourceObjCy >= targetObj->BB.ymin) && (sourceObjCy <= targetObj->BB.ymax) &&
     (sourceObjCz >= targetObj->BB.zmax)) {
    
    if( ABS(sourceObjCz-sourceObj->BB.zmin)*1.2 >= ABS(sourceObjCz-targetObj->BB.zmax)) {
      return ISON;
    }
  }
  
  /* Compute NEXT */
  
  /* Test if sourceObj is next to targetObj */
  /* Condition 1: The Z values of sourceObj and targetObj BB's should intersect = One should not be wholly above the other */
  /* Condition 2: Distance(sourceObj,targetObj) should be less then a constant */
  
  if ( !(sourceObj->BB.zmax <= targetObj->BB.zmin || sourceObj->BB.zmin >= targetObj->BB.zmax) ) {
    if(D2D(sourceObjCx,sourceObjCy,targetObjCx,targetObjCy) < 
       MAX(ABS(sourceObj->BB.xmin-sourceObj->BB.xmax),ABS(sourceObj->BB.ymin-sourceObj->BB.ymax)) &&
       D2D(sourceObjCx,sourceObjCy,targetObjCx,targetObjCy) > 
       MAX(ABS(sourceObj->BB.xmin-sourceObj->BB.xmax)/2,ABS(sourceObj->BB.ymin-sourceObj->BB.ymax)/2)) {
      
      nextTo = TRUE;
      return ISNEXTTO;
    }
  }
  
  if(!nextTo) {
    if ( !(targetObj->BB.zmax <= sourceObj->BB.zmin || targetObj->BB.zmin >= sourceObj->BB.zmax) ) {
      if(D2D(sourceObjCx,sourceObjCy,targetObjCx,targetObjCy) < 
	 MAX(ABS(targetObj->BB.xmin-targetObj->BB.xmax),ABS(targetObj->BB.ymin-targetObj->BB.ymax)) &&
	 D2D(sourceObjCx,sourceObjCy,targetObjCx,targetObjCy) > 
	 MAX(ABS(targetObj->BB.xmin-targetObj->BB.xmax)/2,ABS(targetObj->BB.ymin-targetObj->BB.ymax)/2)) {
	
	nextTo = TRUE;
	return ISNEXTTO;
      }
    }
  }
  
  return NOPLR;  
} 

static SPARK_SPATIAL_RELATION spatial_relation(p3d_rob * object, p3d_rob * robot)
{
  p3d_vector4 targetRealCoord;
  p3d_vector4 targetRelativeCoord;
  p3d_matrix4 inv;
  double rho, phi, theta;
  int isFar;
  double frontAngle = 0.26; //TODO: Make this changeable
  
  if( (robot == NULL) || (object == NULL) ) {
    return NO_RELATION;  
  }
  
  targetRealCoord[0] = object->joints[1]->abs_pos[0][3];
  targetRealCoord[1] = object->joints[1]->abs_pos[1][3];
  targetRealCoord[2] = object->joints[1]->abs_pos[2][3];
  targetRealCoord[3] = 1;
  
  p3d_matInvertXform(robot->joints[1]->abs_pos, inv);
  
  p3d_matvec4Mult(inv, targetRealCoord, targetRelativeCoord);
  
  p3d_cartesian2spherical(targetRelativeCoord[0],targetRelativeCoord[1],targetRelativeCoord[2],
			  &rho, &theta, &phi);
  

  isFar = (SPARK_FAR_LIMIT < D2D(targetRelativeCoord[0],targetRelativeCoord[1],0,0));
  
  //  printf("real coord %f %f %f\n",targetRealCoord[0],targetRealCoord[1],targetRealCoord[2]);
  //  printf("relative coord %f %f %f\n",targetRelativeCoord[0],targetRelativeCoord[1],targetRelativeCoord[2]);
  //  printf("Phi is %f, isFar is %d\n",phi,isFar);
  
  /* Phi is the horizontal one */
  
  if(ABS(phi) < frontAngle) { /* In front */
    if(isFar) return FAR_FRONT ;
    else   return  NEAR_FRONT ;
  }
  if(frontAngle <= phi && phi < 3*M_PI/8) { /* Front left */
    if(isFar)  return  FAR_FRONT_LEFT ;
    else   return NEAR_FRONT_LEFT ;
  }
  if(3*M_PI/8 <= phi && phi < 5*M_PI/8) { /* left */ 
    if(isFar)   return FAR_LEFT ;
    else    return NEAR_LEFT ;
  }
  if(5*M_PI/8 <= phi && phi < 7*M_PI/8) { /* Back left */
    if(isFar)  return  FAR_BACK_LEFT ;
    else   return NEAR_BACK_LEFT ;
  }
  if(7*M_PI/8 <= ABS(phi)) { /* Back */
    if(isFar)  return  FAR_BACK ;
    else   return  NEAR_BACK ;
  }
  if(-7*M_PI/8 <= phi && phi < -5*M_PI/8) { /* Back right */
    if(isFar)  return FAR_BACK_RIGHT ;
    else  return NEAR_BACK_RIGHT ;
  }
  if(-5*M_PI/8 <= phi && phi < -3*M_PI/8) { /* right */
    if(isFar) return FAR_RIGHT ;
    else   return NEAR_RIGHT ;
  }
  if(-3*M_PI/8 <= phi && phi < -1*frontAngle) { /* Front right */
    if(isFar)  return FAR_FRONT_RIGHT ;
    else   return NEAR_FRONT_RIGHT ;
  }
  printf("Bad angle value, This shouldn't happen.\n");
  
  return NO_RELATION;
}

/* sparkUpdateObjectsStatesStart  -  codel START of UpdateObjectsStates
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkUpdateObjectsStatesStart(int *report)
{
  gPrevObjectStates = MY_ALLOC(SPARK_ALL_OBJECT_STATES, 1); /* ALLOC */
  
  return EXEC;
}

/* sparkUpdateObjectsStatesMain  -  codel EXEC of UpdateObjectsStates
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkUpdateObjectsStatesMain(int *report)
{
  SPARK_VISIBILITY objRobotVis, objAgentVis;
  int objRobotVisPl, objAgentVisPl;
  SPARK_REACHABILITY objRobotReach, objHumanReach;
  int pointingResult = FALSE;
  configPt objectQ;
  int i, j, agent_i, agent_j;
  double elevation = 0, azimuth = 0;
  int plRelationIndex = 0;
  SPARK_PLACEMENT_RELATION plrelation;
  HRI_AGENT *current_agent = NULL;

  if(SDI_F->updateObjectsStates == FALSE) {
    return EXEC;
  }

  memcpy(gPrevObjectStates, &SDI_F->objStates, sizeof(SPARK_ALL_OBJECT_STATES));

  for(i=0; i<gEnv->nr; i++) { // For all objects in SPARK
    
    if(!strcmp(SDI_F->objStates.object[i].name.name,"NOT_VALID_OBJECT"))
      continue;

    // find visibility placement for robot
    hri_object_visibility_placement(gAgents->robots[0], gEnv->robot[i], &objRobotVisPl, &elevation, &azimuth);
    
    if ( (SPARK_VISIBILITY_PLACEMENT)objRobotVisPl != OOF ) {
      
      object_visibility_state(gEnv->robot[i], gAgents->robots[0], elevation, azimuth, 50, &objRobotVis);
     
      if( objRobotVis == VISIBLE && !SDI_F->objStates.object[i].detected && !strcasestr(gEnv->robot[i]->name,"TABLE") &&
	  (SDI_F->state.vimanPosterAv || SDI_F->state.sparkObjPosPosterAv) ) {
	// Object has dissappeared	
	objectQ = MY_ALLOC(double, gEnv->robot[i]->nb_dof); /* ALLOC */
	p3d_get_robot_config_into(gEnv->robot[i], &objectQ);
	objectQ[6] = objectQ[7] = objectQ[8] = 0;      
	p3d_set_and_update_this_robot_conf(gEnv->robot[i], objectQ);
	MY_FREE(objectQ, double, gEnv->robot[i]->nb_dof); /* FREE */
	SDI_F->objStates.object[i].disappeared = TRUE;
#ifdef ORO
	updateObject(&SDI_F->objStates.object[i], &gPrevObjectStates->object[i]);
#endif	
	continue; // We pass to the next object
      }
      
      SDI_F->objStates.object[i].disappeared = FALSE;
      strcpy(SDI_F->objStates.object[i].agentVisibilityState[0].agentName.name, gAgents->all_agents[0]->robotPt->name);
      SDI_F->objStates.object[i].agentVisibilityState[0].visibility = objRobotVis;
      SDI_F->objStates.object[i].agentVisibilityState[0].visibilityPlacement = (SPARK_VISIBILITY_PLACEMENT) objRobotVisPl;
      
      SDI_F->objStates.object[i].agentVisibilityState[0].spatialRelation = spatial_relation(gEnv->robot[i], gAgents->all_agents[0]->robotPt);
      
      SDI_F->objStates.object[i].agentVisibilityStateNb = 1;
            
      // Agent Visibility and spatial relations
      
      for(agent_i = 1; agent_i<gAgents->all_agents_no; agent_i++) {
	
	current_agent = gAgents->all_agents[agent_i];
	
	hri_object_visibility_placement(current_agent, gEnv->robot[i], &objAgentVisPl, &elevation, &azimuth);
	object_visibility_state(gEnv->robot[i], current_agent, elevation, azimuth, 50, &objAgentVis);
		
	strcpy(SDI_F->objStates.object[i].agentVisibilityState[agent_i].agentName.name,current_agent->robotPt->name);
	SDI_F->objStates.object[i].agentVisibilityState[agent_i].visibility = objAgentVis;
	SDI_F->objStates.object[i].agentVisibilityState[agent_i].visibilityPlacement = (SPARK_VISIBILITY_PLACEMENT) objAgentVisPl;
	
	// -----Pointing
	hri_object_pointing_placement(current_agent, gEnv->robot[i], &objAgentVisPl, &elevation, &azimuth);
	if((SPARK_VISIBILITY_PLACEMENT) objAgentVisPl == FOA) {
	  pointingResult = hri_is_object_pointed(current_agent,  gEnv->robot[i], 50, FALSE);
	  if(pointingResult == TRUE)
	    SDI_F->objStates.object[i].agentVisibilityState[agent_i].pointedAt = TRUE;
	  else
	    SDI_F->objStates.object[i].agentVisibilityState[agent_i].pointedAt = FALSE;
	}
	else {
	  SDI_F->objStates.object[i].agentVisibilityState[agent_i].pointedAt = FALSE;
	}
	// -----
	SDI_F->objStates.object[i].agentVisibilityState[agent_i].spatialRelation = spatial_relation(gEnv->robot[i], current_agent->robotPt);
	
	SDI_F->objStates.object[i].agentVisibilityStateNb++;
      }

      // Reachability computation is transfered to computeAllRechabilities Request
      //objRobotReach = spark_object_reachability_state(gEnv->robot[i], gAgents->robots[0]);
      //SDI_F->objStates.object[i].robotReachability = objRobotReach;
      
      //objHumanReach = spark_object_reachability_state(gEnv->robot[i], gAgents->humans[0]);
      //SDI_F->objStates.object[i].humanReachability = objHumanReach;
      
      SDI_F->objStates.object[i].pos.x = gEnv->robot[i]->joints[1]->abs_pos[0][3];
      SDI_F->objStates.object[i].pos.y = gEnv->robot[i]->joints[1]->abs_pos[1][3];
      SDI_F->objStates.object[i].pos.z = gEnv->robot[i]->joints[1]->abs_pos[2][3];

      // Compute Object spatial placement
      plRelationIndex = 0;
      for(j=0; j<gEnv->nr; j++) { // For all objects in SPARK
	if(i!=j && ( SDI_F->objStates.object[j].detected || strcasestr(gEnv->robot[j]->name,"TABLE") ) ) {
	  // we compute relative placements only with detected objects
	  plrelation =  placement_relation(gEnv->robot[i],gEnv->robot[j]);
	  
	  if(plrelation != NOPLR) {
	    SDI_F->objStates.object[i].placements[plRelationIndex].relation = plrelation;
	    strcpy(SDI_F->objStates.object[i].placements[plRelationIndex].object.name, gEnv->robot[j]->name);
	    plRelationIndex++;
	  }
	}
      }

      SDI_F->objStates.object[i].placementLength = plRelationIndex; 

      //printf("Info computed for %s - %d\n", gEnv->robot[i]->name, time(0));

#ifdef ORO
      updateObject(&SDI_F->objStates.object[i], &gPrevObjectStates->object[i] );
#endif
    }
  }
#ifdef ORO 
  // Agent related stuff 
  for(agent_i = 0; agent_i<gAgents->all_agents_no; agent_i++) {
   
    for(agent_j = 0; agent_j<gAgents->all_agents_no; agent_j++) {

      updateAgentPosture(gAgents->all_agents[agent_i], agent_i, (SPARK_HUMAN_STATE_ENUM)gAgents->all_agents[agent_i]->actual_state);
      updateAgentPosture(gAgents->all_agents[agent_i], agent_j, (SPARK_HUMAN_STATE_ENUM)gAgents->all_agents[agent_i]->actual_state);

      if(agent_i != agent_j) {
	updateAgentSp(agent_i, gAgents->all_agents[agent_j], spatial_relation(gAgents->all_agents[agent_j]->robotPt, gAgents->all_agents[agent_i]->robotPt));
	
	hri_object_visibility_placement(gAgents->all_agents[agent_i], gAgents->all_agents[agent_j]->robotPt, &objAgentVisPl, &elevation, &azimuth);
	object_visibility_state(gAgents->all_agents[agent_j]->robotPt,gAgents->all_agents[agent_i] , elevation, azimuth, 50, &objAgentVis);	
	
	updateAgentAgentVisibility(gAgents->all_agents[agent_i],agent_i, gAgents->all_agents[agent_j],
				   (SPARK_VISIBILITY_PLACEMENT)objAgentVisPl, objAgentVis);
	updateAgentAgentVisibility(gAgents->all_agents[agent_i],agent_j, gAgents->all_agents[agent_j],
				   (SPARK_VISIBILITY_PLACEMENT)objAgentVisPl, objAgentVis);
      }
    }
  }
#endif
  
  sparkObjectStatesobjStatesPosterWrite(SPARK_OBJECTSTATES_POSTER_ID, &SDI_F->objStates);

#ifdef ORO  
  sendToORO(SDI_F->oroServerAddress.address, SDI_F->oroServerAddress.port);
#endif
  
  SDI_F->updateObjectsStates = FALSE;
  
  return EXEC;
}

/* sparkUpdateObjectsStatesEnd  -  codel END of UpdateObjectsStates
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkUpdateObjectsStatesEnd(int *report)
{
  MY_FREE(gPrevObjectStates, SPARK_ALL_OBJECT_STATES, 1); /* ALLOC */
  printf("Object State Update is stopped\n");
  *report = OK;
  return ETHER;
}

/* sparkUpdateObjectsStatesInter -  codel INTER of UpdateObjectsStates
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkUpdateObjectsStatesInter(int *report)
{
  MY_FREE(gPrevObjectStates, SPARK_ALL_OBJECT_STATES, 1); /* ALLOC */
  printf("Object State Update is stopped\n");
  *report = OK;
  return ETHER;
}

/* sparkComputeAllReachabilitiesMain  -  codel EXEC of ComputeAllReachabilities
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkComputeAllReachabilitiesMain(int *report)
{
  int i,j;
  int agentReachIdx = 0;
  
  if(gAllReachs == NULL) {
    gAllReachs = MY_ALLOC(SPARK_ALL_REACHABILITIES, 1);
    gAllReachs->agentReach = MY_ALLOC(SPARK_AGENT_REACHABILITY, gAgents->all_agents_no * (gEnv->nr - 1));
    gAllReachs->agentReachNb = gAgents->all_agents_no * (gEnv->nr - 1);
  }
  
  for(i=0; i<gAgents->all_agents_no; i++) {
    for(j=0; j<gEnv->nr; j++) {

      // No need to check self reachability :-)
      if(gEnv->robot[j] != gAgents->all_agents[i]->robotPt) {
	
	// Check if j'th object is a valid one
	if(strcmp(SDI_F->objStates.object[j].name.name,"NOT_VALID_OBJECT")) {
	  
	  strcpy(gAllReachs->agentReach[agentReachIdx].agentName.name,
		 gAgents->all_agents[i]->robotPt->name);
	  strcpy(gAllReachs->agentReach[agentReachIdx].objectName.name,
		 gEnv->robot[j]->name);
	  if(!SDI_F->objStates.object[j].exists || SDI_F->objStates.object[j].disappeared) 
	    gAllReachs->agentReach[agentReachIdx].reachability = REACHABILITY_UNKNOWN;
	  else
	    gAllReachs->agentReach[agentReachIdx].reachability = spark_object_reachability_state(gEnv->robot[j], gAgents->all_agents[i]);		      
	  
	  printf("Agent Name: %s Object name: %s REACHABILITY: %d\n", 
		 gAllReachs->agentReach[agentReachIdx].agentName.name,
		 gAllReachs->agentReach[agentReachIdx].objectName.name,
		 gAllReachs->agentReach[agentReachIdx].reachability);
	  
	  gAllReachs->agentReachNb = ++agentReachIdx;
	}
      }
    }
  }
  
#ifdef ORO
  sendReachabilitiesToORO(gAllReachs, SDI_F->oroServerAddress.address, SDI_F->oroServerAddress.port);
#endif 
  
  return ETHER;
}


/* sparkComputeFactsStart  -  codel START of ComputeFacts
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkComputeFactsStart(int *report)
{  
  return EXEC;
}


/* sparkComputeFactsMain  -  codel EXEC of ComputeFacts
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkComputeFactsMain(int *report)
{ 
  float clock_val = 0;
  float elapsed_time = 0;
  
  //clock_val = clock();
  
  hri_compute_geometric_facts(gAgents, gEntities);

#ifdef ORO
  prepareFactsForORO(gAgents, gEntities);
  sendToORO(SDI_F->oroServerAddress.address, SDI_F->oroServerAddress.port);
#endif
  
  //elapsed_time = (clock() - clock_val)/CLOCKS_PER_SEC;
  //printf("TIME passed for: %f overall,\n", elapsed_time);
  
  return ETHER;
}


/* sparkComputeFactsEnd  -  codel END of ComputeFacts
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkComputeFactsEnd(int *report)
{
  return ETHER;
}


/* sparkComputeFactsInter  -  codel INTER of ComputeFacts
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkComputeFactsInter(int *report)
{
  return ETHER;
}

